# import sys
# import os

# sys.path.append(os.path.dirname(os.path.abspath(__file__)))

# if sys.platform == "darwin":
#     import jacspy
# elif sys.platform == "linux":
#     from linux import jacspylinux as jacspy
